/*
 * Wine Driver for PulseAudio - DSound Output Functionality
 * http://pulseaudio.org/
 *
 * Copyright	2009 Arthur Talyor <theycallhimart@gmail.com>
 *
 * Conatins code from other wine multimedia drivers.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "config.h"

#include <stdarg.h>

#include "windef.h"
#include "winbase.h"
#include "wingdi.h"
#include "winuser.h"
#include "winerror.h"
#include "mmddk.h"
#include "dsound.h"
#include "dsdriver.h"
#include "winreg.h"

#include <winepulse.h>
#include "wine/debug.h"

WINE_DEFAULT_DEBUG_CHANNEL(dspulse);

#if HAVE_PULSEAUDIO

/*======================================================================*
 *                  Low level DSOUND implementation			*
 *======================================================================*/

/* A buffer is allocated with a pointer indicating the read position in the
 * buffer. The pulse write callback reads data from the buffer, updating the
 * read pointer to the location at the end of the read. Upon reaching the end
 * or attempting to read past the end of buffer the read pointer wraps around
 * to the beginning again. DirectSound applications can write anywhere in the
 * buffer at anytime without locking and can know the location of the read
 * pointer. The position of the read pointer cannot be changed by the
 * application and access to it uses a locking scheme. A fake pointer
 * indicating estimated playback position is also available to the application.
 * Applications can potentially write to the same area of memory which is also
 * being read by the pulse thread. However, this is uncommon as directsound
 * applications know where pulse should be reading from via the pointer
 * locations and MSDN says that such an operation should be avoided with the
 * results being undefined.
 */

/* Fragment lengths try to be a power of two close to 10ms worth of data. See
 * dlls/dsound/mixer.c
 */
static int fragment_length(pa_sample_spec *s) {
    if (s->rate <= 12800)
        return 128 * pa_frame_size(s);

    if (s->rate <= 25600)
        return 256 * pa_frame_size(s);

    if (s->rate <= 51200)
        return 512 * pa_frame_size(s);

    return 1024 * pa_frame_size(s);
}

/* Callback from the pulse thread for stream data */
static void DSPULSE_BufferReadCallback(pa_stream *s, size_t nbytes, void *userdata) {
    IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)userdata;

    if (!This->buffer)
        return;

    /* Fraglens are always powers of 2 */
    nbytes+= This->fraglen - 1;
    nbytes&= ~(This->fraglen - 1);

    /* If we advance more than 10 fragments at a time it appears that the buffer
     * pointer is never advancing because of wrap-around. Evil magic numbers. */
    if (nbytes > This->fraglen * 5)
        nbytes = This->fraglen * 5;

    TRACE("Reading %u bytes.\n", nbytes);

    if (This->buffer_read_offset + nbytes <= This->buffer_length) {
        pa_stream_write(s, This->buffer + This->buffer_read_offset, nbytes, NULL, 0, PA_SEEK_RELATIVE);
        This->buffer_play_offset = This->buffer_read_offset;
        This->buffer_read_offset += nbytes;
    } else {
        size_t write_length = This->buffer_length - This->buffer_read_offset;
        nbytes -= write_length;
        pa_stream_write(s, This->buffer + This->buffer_read_offset, write_length, NULL, 0, PA_SEEK_RELATIVE);
        pa_stream_write(s, This->buffer, nbytes, NULL, 0, PA_SEEK_RELATIVE);
        This->buffer_play_offset = This->buffer_read_offset;
        This->buffer_read_offset = nbytes;
    }

    This->buffer_read_offset %= This->buffer_length;
}

/* Called when the stream underruns. Just for information */
static void DSPULSE_BufferUnderflowCallback(pa_stream *s, void *userdata) {
    WARN("(%p) underrun.\n", userdata);
}

/* Connects a stream to the server. Does not update
 * IDsDriverBufferImpl->fraglen. Does not lock the pulse mainloop or free
 * objects in case of failure. This should be handled by the calling function.
 */
static HRESULT DSPULSE_ConnectStream(IDsDriverBufferImpl* This) {
    pa_buffer_attr ba_request;
    const pa_buffer_attr *ba_obtained;
    char c[PA_SAMPLE_SPEC_SNPRINT_MAX];
    pa_stream_flags_t stream_flags = PA_STREAM_START_CORKED;

#if PA_PROTOCOL_VERSION >= 14
    /* We are a "fragment wait based" application, so this flag should be
     * acceptable. */
    stream_flags |= PA_STREAM_EARLY_REQUESTS;
#elif PA_PROTOCOL_VERSION >= 13 && PA_API_VERSION >= 12
    stream_flags |= PA_STREAM_ADJUST_LATENCY;
#endif

    pa_sample_spec_snprint(c, PA_SAMPLE_SPEC_SNPRINT_MAX, &This->sample_spec);
    TRACE("Sample spec %s fragment size %u.\n", c, This->fraglen);

    ba_request.tlength = This->fraglen * 4;     //  ~40ms
    ba_request.minreq = This->fraglen;          //  ~10ms
    ba_request.prebuf = (uint32_t)-1;           //  same as tlength
    ba_request.maxlength = This->buffer_length; //  2^x = ~3s

    TRACE("Asking for buffer tlength:%u (%llums) minreq:%u (%llums)\n",
        ba_request.tlength, pa_bytes_to_usec(ba_request.tlength, &This->sample_spec)/1000,
        ba_request.minreq, pa_bytes_to_usec(ba_request.minreq, &This->sample_spec)/1000);

    This->stream = pa_stream_new(PULSE_context, "DirectSound Buffer", &This->sample_spec, NULL);
    if (!This->stream) return DSERR_BADFORMAT;

    pa_stream_set_state_callback(This->stream, PULSE_StreamStateCallback, This);
    pa_stream_set_write_callback(This->stream, DSPULSE_BufferReadCallback, This);
    pa_stream_set_underflow_callback(This->stream, DSPULSE_BufferUnderflowCallback, This);

    TRACE("Attempting to connect (%p)->stream for playback on %s\n", This, WOutDev[This->drv->wDevID].device_name);
    pa_stream_connect_playback(This->stream, WOutDev[This->drv->wDevID].device_name, &ba_request, stream_flags, NULL, NULL);
    for (;;) {
        pa_context_state_t cstate = pa_context_get_state(PULSE_context);
        pa_stream_state_t sstate = pa_stream_get_state(This->stream);

        if (cstate == PA_CONTEXT_FAILED || cstate == PA_CONTEXT_TERMINATED ||
            sstate == PA_STREAM_FAILED || sstate == PA_STREAM_TERMINATED) {
            ERR("Failed to connect stream context object: %s\n", pa_strerror(pa_context_errno(PULSE_context)));
            return DSERR_BUFFERLOST;
        }

        if (sstate == PA_STREAM_READY)
            break;

        pa_threaded_mainloop_wait(PULSE_ml);
    }
    TRACE("(%p)->stream connected for playback.\n", This);
    ba_obtained = pa_stream_get_buffer_attr(This->stream);

    TRACE("Obtained buffer tlength:%u (%llums) minreq:%u (%llums)\n",
        ba_obtained->tlength, pa_bytes_to_usec(ba_obtained->tlength, &This->sample_spec)/1000,
        ba_obtained->minreq, pa_bytes_to_usec(ba_obtained->minreq, &This->sample_spec)/1000);

    return DS_OK;
}

static HRESULT WINAPI IDsDriverBufferImpl_QueryInterface(PIDSDRIVERBUFFER iface, REFIID riid, LPVOID *ppobj) {
    /* IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)iface; */
    FIXME("(): stub!\n");
    return DSERR_UNSUPPORTED;
}

static ULONG WINAPI IDsDriverBufferImpl_AddRef(PIDSDRIVERBUFFER iface) {
    IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)iface;
    ULONG refCount = InterlockedIncrement(&This->ref);

    TRACE("(%p)->(ref before=%u)\n",This, refCount - 1);

    return refCount;
}

static ULONG WINAPI IDsDriverBufferImpl_Release(PIDSDRIVERBUFFER iface) {
    IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)iface;
    ULONG refCount = InterlockedDecrement(&This->ref);

    TRACE("(%p)->(ref before=%u)\n",This, refCount + 1);

    if (refCount)
        return refCount;

    TRACE("mmap buffer %p destroyed\n", This->buffer);
    pa_threaded_mainloop_lock(PULSE_ml);
    PULSE_WaitForOperation(pa_stream_cork(This->stream, 1, PULSE_StreamSuccessCallback, This));
    pa_stream_disconnect(This->stream);
    if (This == This->drv->primary)
        This->drv->primary = NULL;
    HeapFree(GetProcessHeap(), 0, This->buffer);
    This->buffer = NULL;
    pa_stream_unref(This->stream);
    This->stream = NULL;
    HeapFree(GetProcessHeap(), 0, This);
    pa_threaded_mainloop_unlock(PULSE_ml);

    return 0;
}

static HRESULT WINAPI IDsDriverBufferImpl_Lock(PIDSDRIVERBUFFER iface,
					       LPVOID*ppvAudio1,LPDWORD pdwLen1,
					       LPVOID*ppvAudio2,LPDWORD pdwLen2,
					       DWORD dwWritePosition,DWORD dwWriteLen,
					       DWORD dwFlags)
{
    /* We set DSDDESC_DONTNEEDPRIMARYLOCK so this should never be called */
    TRACE("(%p): stub", iface);
    return DSERR_UNSUPPORTED;
}

static HRESULT WINAPI IDsDriverBufferImpl_Unlock(PIDSDRIVERBUFFER iface,
						 LPVOID pvAudio1,DWORD dwLen1,
						 LPVOID pvAudio2,DWORD dwLen2)
{
    /* We set DSDDESC_DONTNEEDPRIMARYLOCK so this should never be called */
    TRACE("(%p): stub", iface);
    return DSERR_UNSUPPORTED;
}

/* You cannot change the sample format of a connected stream, so we need to
 * destroy and re-create the stream if the sample spec is different */
static HRESULT WINAPI IDsDriverBufferImpl_SetFormat(PIDSDRIVERBUFFER iface, LPWAVEFORMATEX pwfx) {
    IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)iface;
    pa_sample_spec old_spec;

    TRACE("(%p, %p)\n", iface, pwfx);

    old_spec.rate = This->sample_spec.rate;
    old_spec.format = This->sample_spec.format;
    old_spec.channels = This->sample_spec.channels;

    if (!PULSE_SetupFormat(pwfx, &This->sample_spec))
        return DSERR_BADFORMAT;

    if (old_spec.rate ==    This->sample_spec.rate &&
        old_spec.format ==  This->sample_spec.format &&
        old_spec.channels == This->sample_spec.channels) {
        TRACE("same as original sample spec, exiting.\n");
        return DS_OK;
    }

    /* If the format doesn't match, return an error and the buffer will be remade */
    TRACE("Formats don't match, failing causing re-creation.\n");
    return DSERR_BUFFERLOST;
}

static HRESULT WINAPI IDsDriverBufferImpl_SetFrequency(PIDSDRIVERBUFFER iface, DWORD dwFreq)
{
    /* IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)iface; */
    /* You can't do for primary buffers anyways */
    return S_OK;
}

static HRESULT WINAPI IDsDriverBufferImpl_SetVolumePan(PIDSDRIVERBUFFER iface, PDSVOLUMEPAN pVolPan) {
    IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)iface;

    if (This->sample_spec.channels == 2) {
        This->volume.channels = 2;
        if (pVolPan->lPan < 0) {
            This->volume.values[0] = pa_sw_volume_from_dB(pVolPan->lVolume/-10.0 + pVolPan->lPan/10.0);
            This->volume.values[1] = pa_sw_volume_from_dB(pVolPan->lVolume/-10.0);
        } else {
            This->volume.values[0] = pa_sw_volume_from_dB(pVolPan->lVolume/-10.0);
            This->volume.values[1] = pa_sw_volume_from_dB(pVolPan->lVolume/-10.0 - pVolPan->lPan/10.0);
        }
    } else {
        WARN("Panning non-stereo streams not supported yet!\n");
        pa_cvolume_set(&This->volume, This->sample_spec.channels, pa_sw_volume_from_dB(pVolPan->lVolume/-10.0));
        /* Would be nice to return DSERR_CONTROLUNAVAIL, but that isn't up to us
         * here */
    }

    pa_threaded_mainloop_lock(PULSE_ml);
    if (This->stream && PULSE_context && pa_context_get_state(PULSE_context) == PA_CONTEXT_READY &&
	pa_stream_get_state(This->stream) == PA_STREAM_READY && pa_cvolume_valid(&This->volume))
        PULSE_WaitForOperation(
	    pa_context_set_sink_input_volume(
		PULSE_context, pa_stream_get_index(This->stream),
		&This->volume, PULSE_ContextSuccessCallback, This));

    pa_threaded_mainloop_unlock(PULSE_ml);

    return DS_OK;
}

static HRESULT WINAPI IDsDriverBufferImpl_SetPosition(PIDSDRIVERBUFFER iface, DWORD dwNewPos)
{
    /* IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)iface; */
    /* Not allowed on primary buffers */
    return DSERR_UNSUPPORTED;
}

static HRESULT WINAPI IDsDriverBufferImpl_GetPosition(PIDSDRIVERBUFFER iface,
						      LPDWORD lpdwPlay, LPDWORD lpdwWrite)
{
    IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)iface;

    pa_threaded_mainloop_lock(PULSE_ml);
    if (!This->buffer || pa_stream_get_state(This->stream) != PA_STREAM_READY) {
        pa_threaded_mainloop_unlock(PULSE_ml);
        return DSERR_UNINITIALIZED;
    }

    if (lpdwPlay)
        *lpdwPlay = This->buffer_play_offset;
    if (lpdwWrite)
        *lpdwWrite = This->buffer_read_offset;
    pa_threaded_mainloop_unlock(PULSE_ml);

    return DS_OK;
}

static HRESULT WINAPI IDsDriverBufferImpl_Play(PIDSDRIVERBUFFER iface, DWORD dwRes1, DWORD dwRes2, DWORD dwFlags) {
    IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)iface;
    TRACE("(%p,%x,%x,%x)\n",iface,dwRes1,dwRes2,dwFlags);

    pa_threaded_mainloop_lock(PULSE_ml);
    PULSE_WaitForOperation(pa_stream_cork(This->stream, 0, PULSE_StreamSuccessCallback, This));
    pa_threaded_mainloop_unlock(PULSE_ml);
    return DS_OK;
}

static HRESULT WINAPI IDsDriverBufferImpl_Stop(PIDSDRIVERBUFFER iface) {
    IDsDriverBufferImpl *This = (IDsDriverBufferImpl *)iface;
    TRACE("(%p)\n",iface);

    pa_threaded_mainloop_lock(PULSE_ml);
    PULSE_WaitForOperation(pa_stream_cork(This->stream, 1, PULSE_StreamSuccessCallback, This));
    pa_threaded_mainloop_unlock(PULSE_ml);
    return DS_OK;
}

/*****************************************************************************/

static const IDsDriverBufferVtbl dsdbvt =
{
    IDsDriverBufferImpl_QueryInterface,
    IDsDriverBufferImpl_AddRef,
    IDsDriverBufferImpl_Release,
    IDsDriverBufferImpl_Lock,
    IDsDriverBufferImpl_Unlock,
    IDsDriverBufferImpl_SetFormat,
    IDsDriverBufferImpl_SetFrequency,
    IDsDriverBufferImpl_SetVolumePan,
    IDsDriverBufferImpl_SetPosition,
    IDsDriverBufferImpl_GetPosition,
    IDsDriverBufferImpl_Play,
    IDsDriverBufferImpl_Stop
};

static HRESULT WINAPI IDsDriverImpl_CreateSoundBuffer(PIDSDRIVER iface,
						      LPWAVEFORMATEX pwfx,
						      DWORD dwFlags, DWORD dwCardAddress,
						      LPDWORD pdwcbBufferSize,
						      LPBYTE *ppbBuffer,
						      LPVOID *ppvObj) {
    IDsDriverImpl *This = (IDsDriverImpl *)iface;
    IDsDriverBufferImpl** ippdsdb = (IDsDriverBufferImpl**)ppvObj;
    IDsDriverBufferImpl *That;
    HRESULT ret;

    TRACE("(%p,%p,%x,%x)\n",iface,pwfx,dwFlags,dwCardAddress);
    /* we only support primary buffers */
    
    if (!(dwFlags & DSBCAPS_PRIMARYBUFFER))
        return DSERR_UNSUPPORTED;
    if (This->primary)
        return DSERR_ALLOCATED;
    This->primary = *ippdsdb;

    *ippdsdb = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(IDsDriverBufferImpl));

    if (!*ippdsdb) {
        ERR("Out of memory\n");
        return DSERR_OUTOFMEMORY;
    }

    That = *ippdsdb;
    That->lpVtbl  = &dsdbvt;
    That->ref = 1;

    That->drv = This;
    TRACE("IdsDriverBufferImpl %p created.\n", That);
        
    if (!PULSE_SetupFormat(pwfx, &That->sample_spec)) {
        WARN("Bad audio format.\n");
        ret = DSERR_BADFORMAT;
        goto err;
    }

    /* The buffer length has to be greater than fraglen * 20 or else logic in
     * dlls/dsound/mixer.c fails to correctly understand buffer wrap around. */
    That->fraglen = fragment_length(&That->sample_spec);
    That->buffer_length = That->fraglen * 32;
    That->buffer = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, That->buffer_length);
    if (!That->buffer) {
        ret = DSERR_OUTOFMEMORY;
        goto err;
    }
    That->buffer_read_offset = 0;
    That->buffer_play_offset = 0;

    pa_threaded_mainloop_lock(PULSE_ml);
    ret = DSPULSE_ConnectStream(That);
    pa_threaded_mainloop_unlock(PULSE_ml);

    if (ret != DS_OK)
        goto err;

    *pdwcbBufferSize = That->buffer_length;
    *ppbBuffer = That->buffer;
    
    TRACE("Exiting with success. Have buffer %p of length %u\n", That->buffer, That->buffer_length);
    return DS_OK;
    
err:
    pa_threaded_mainloop_lock(PULSE_ml);
    if (That->stream) {
	    if (pa_stream_get_state(That->stream) == PA_STREAM_READY)
	        pa_stream_disconnect(That->stream);
	    pa_stream_unref(That->stream);
	    That->stream = NULL;
    }
    pa_threaded_mainloop_unlock(PULSE_ml);
    HeapFree(GetProcessHeap(), 0, That->buffer);
    HeapFree(GetProcessHeap(), 0, *ippdsdb);
    WARN("exiting with failure.\n");
    return ret;
}

static HRESULT WINAPI IDsDriverImpl_QueryInterface(PIDSDRIVER iface, REFIID riid, LPVOID *ppobj) {
    /* IDsDriverImpl *This = (IDsDriverImpl *)iface; */
    FIXME("(%p): stub!\n",iface);
    return DSERR_UNSUPPORTED;
}

static ULONG WINAPI IDsDriverImpl_AddRef(PIDSDRIVER iface) {
    IDsDriverImpl *This = (IDsDriverImpl *)iface;
    ULONG refCount = InterlockedIncrement(&This->ref);

    TRACE("(%p)->(ref before=%u)\n",This, refCount - 1);

    return refCount;
}

static ULONG WINAPI IDsDriverImpl_Release(PIDSDRIVER iface) {
    IDsDriverImpl *This = (IDsDriverImpl *)iface;
    ULONG refCount = InterlockedDecrement(&This->ref);

    TRACE("(%p)->(ref before=%u)\n",This, refCount + 1);

    if (refCount)
        return refCount;

    HeapFree(GetProcessHeap(), 0, This);
    return 0;
}

static HRESULT WINAPI IDsDriverImpl_GetDriverDesc(PIDSDRIVER iface, PDSDRIVERDESC pDesc) {
    IDsDriverImpl *This = (IDsDriverImpl *)iface;
    TRACE("(%p,%p)\n",iface,pDesc);
    *pDesc			= WOutDev[This->wDevID].ds_desc;
    pDesc->dwFlags		= DSDDESC_DONTNEEDSECONDARYLOCK | DSDDESC_DONTNEEDPRIMARYLOCK;
    pDesc->dnDevNode		= 0; /*TODO: Bwah? */
    pDesc->wVxdId           = 0;
    pDesc->wReserved		= 0;
    pDesc->ulDeviceNum		= This->wDevID;
    pDesc->dwHeapType		= DSDHEAP_NOHEAP;
    pDesc->pvDirectDrawHeap	= NULL;
    pDesc->dwMemStartAddress	= 0xDEAD0000;
    pDesc->dwMemEndAddress	= 0xDEAF0000;
    pDesc->dwMemAllocExtra	= 0;
    pDesc->pvReserved1		= NULL;
    pDesc->pvReserved2		= NULL;
    return DS_OK;
}

static HRESULT WINAPI IDsDriverImpl_Close(PIDSDRIVER iface) {
    IDsDriverImpl *This = (IDsDriverImpl *)iface;
    TRACE("(%p) stub, harmless\n",This);
    return DS_OK;
}

static HRESULT WINAPI IDsDriverImpl_Open(PIDSDRIVER iface) {
    IDsDriverImpl *This = (IDsDriverImpl *)iface;
    TRACE("(%p) stub, harmless\n",This);
    return S_OK;
}

static HRESULT WINAPI IDsDriverImpl_GetCaps(PIDSDRIVER iface, PDSDRIVERCAPS pCaps) {
    IDsDriverImpl *This = (IDsDriverImpl *)iface;

    if (pa_context_get_state(PULSE_context) != PA_CONTEXT_READY) {
        ERR("Context failure.\n");
        return DSERR_GENERIC;
    }

    TRACE("(%p,%p)\n",iface,pCaps);
    *pCaps = WOutDev[This->wDevID].ds_caps;
    return DS_OK;
}

static HRESULT WINAPI IDsDriverImpl_DuplicateSoundBuffer(PIDSDRIVER iface,
							 PIDSDRIVERBUFFER pBuffer,
							 LPVOID *ppvObj)
{
    IDsDriverImpl *This = (IDsDriverImpl *)iface;
    FIXME("(%p,%p): stub\n",This,pBuffer);
    return DSERR_INVALIDCALL;
}

static const IDsDriverVtbl dsdvt =
{
    IDsDriverImpl_QueryInterface,
    IDsDriverImpl_AddRef,
    IDsDriverImpl_Release,
    IDsDriverImpl_GetDriverDesc,
    IDsDriverImpl_Open,
    IDsDriverImpl_Close,
    IDsDriverImpl_GetCaps,
    IDsDriverImpl_CreateSoundBuffer,
    IDsDriverImpl_DuplicateSoundBuffer
};

DWORD wodDsCreate(UINT wDevID, PIDSDRIVER* drv) {
    IDsDriverImpl** idrv = (IDsDriverImpl**)drv;

    if (pa_context_get_state(PULSE_context) != PA_CONTEXT_READY || pa_context_is_local(PULSE_context) != 1) {
        WARN("Connection failure or server is not local, falling back to WaveOut HEL.\n");
        return MMSYSERR_NOTSUPPORTED;
    }

    *idrv = HeapAlloc(GetProcessHeap(), 0, sizeof(IDsDriverImpl));
    if (!*idrv)
        return MMSYSERR_NOMEM;
    
    TRACE("IDsDriverImpl %p created.\n", *idrv);

    (*idrv)->lpVtbl = &dsdvt;
    (*idrv)->ref = 1;
    (*idrv)->wDevID = wDevID;
    (*idrv)->primary = NULL;

    return MMSYSERR_NOERROR;
}

DWORD wodDsDesc(UINT wDevID, PDSDRIVERDESC desc) {
    TRACE("(%u, %p)\n", wDevID, desc);
    *desc = WOutDev[wDevID].ds_desc;
    return MMSYSERR_NOERROR;
}
#endif /* HAVE_PULSEAUDIO */
